//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                        Real Time Clock (RTC) module                        *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Emulates the MC146818 RTC circuit.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.0.0 - 15 October 2007, S.J.Kay
// - Created a new file and implement the RTC emulation.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

//#ifdef MINGW
//#include <windows.h>
//#else
//#endif

#include "rtc.h"
#include "ubee512.h"
#include "mz80.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
static UINT8 addr;

static rtc_u rtc;

static char name[512];

extern int rtc_use;

extern uint64_t z80_cycles;

extern int model;
extern char userhome_rtcpath[];
extern char *model_args[];

extern FILE *debuglog;
extern int debuglvl;

//==============================================================================
// Set clock times from host. 
//
// struct tm
//
// This structure is used to hold the values representing a calendar time.
// It contains the following members, with the meanings as shown.
//
// int tm_sec      /* seconds after minute [0-61] (61 allows for 2 leap-seconds)*/
// int tm_min      /* minutes after hour [0-59] */
// int tm_hour     /* hours after midnight [0-23] */
// int tm_mday     /* day of the month [1-31] */
// int tm_mon      /* month of year [0-11] */
// int tm_year     /* current year-1900 */
// int tm_wday     /* days since Sunday [0-6] */
// int tm_yday     /* days since January 1st [0-365] */
// int tm_isdst    /* daylight savings indicator */
//
//   pass: void
// return: void
//==============================================================================
static void rtc_setclockfromhost (void)
{
 typedef struct tm tm_t;

 time_t result;
 tm_t resultp;
 
 time(&result); 
#ifdef MINGW    
 memcpy(&resultp, localtime(&result), sizeof(resultp));
#else
 localtime_r(&result, &resultp);
#endif 
 if (resultp.tm_sec > 59)		// 2 second leap-seconds ignore
    resultp.tm_sec = 59;
 rtc.member.seconds = resultp.tm_sec;
 rtc.member.minutes = resultp.tm_min;                  
 rtc.member.hours = resultp.tm_hour;
 rtc.member.mday = resultp.tm_mday;
 rtc.member.month = resultp.tm_mon + 1;
 rtc.member.year = resultp.tm_year - 100;     
 rtc.member.wday = resultp.tm_wday + 1;          
}

//==============================================================================
// RTC Initialise.
//
// Load the RTC file for the model being emulated if one exists.
//
//   pass: void
// return: int				0 if success, -1 if error
//==============================================================================
int rtc_init (void)
{
 FILE *fp;
 long len;

 if (rtc_use)
    {
     snprintf(name, sizeof(name), "%s%s.rtc", userhome_rtcpath, model_args[model]);
     fp = fopen(name, "rb");

     if (fp != NULL)
        {
         fseek(fp, 0, SEEK_END);  	        // get size of the RTC file
         len = ftell(fp);
         fseek(fp, 0, SEEK_SET);		// back to start

         if (len != sizeof(rtc))
            {
             console_output();
             printf("rtc_init: RTC file is incorrect size: %s\n", name);
             fclose(fp);
             return -1;
            }
         else
            fread(&rtc, 1, sizeof(rtc), fp);
         fclose(fp);
        }
     else
        memset(&rtc, 0, sizeof(rtc));
        
     rtc.member.reg_d = RTC_D_VRT;
     addr = 0;
     rtc_setclockfromhost();
    }
 return 0;
}

//==============================================================================
// RTC de-initialise.
//
// For rtc out the WAV file header is filled in and the file closed.
// For rtc in the file is closed.
//
//   pass: void
// return: int				0
//==============================================================================
int rtc_deinit (void)
{
 FILE *fp;

 if (rtc_use)
    {
     snprintf(name, sizeof(name), "%s%s.rtc", userhome_rtcpath, model_args[model]);
     fp = fopen(name, "wb");

     if (fp != NULL)
        {
         fwrite(&rtc, 1, sizeof(rtc), fp);
         fclose(fp);
         return 0;
        }
     else
        {
         console_output();
         printf("rtc_deinit: Unable to create RTC file: %s\n", name);
         return -1;
        }
    }
 return 0;
}

//==============================================================================
// RTC reset.
//
//   pass: void
// return: int				0
//==============================================================================
int rtc_reset (void)
{
 if (rtc_use)
    {
     rtc.member.reg_b &= (0xff ^ (RTC_B_PIE | RTC_B_AIE | RTC_B_UIE | RTC_B_SQWE));
     rtc.member.reg_c = 0;
    } 
 return 0;
}

//==============================================================================
// RTC read register data - Port function
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16 register 		register data
//==============================================================================
UINT16 rtc_r (UINT16 port, struct z80PortRead *port_s)
{
 uint8_t data;
 
 if (rtc_use)
    {
#if 0
     if ((standard != 0) && (port > 0x07))
        return 0;
#endif
     switch (port & 0x0f)
        {
         case 0x04 : if (debuglvl > 0)
                        fprintf(debuglog, "rtc_r: port=0x%02X data=0x%02X\n", port, addr);
//                     printf("rtc_r: port=0x%02X data=0x%02X\n", port, addr);
                     return addr;
         case 0x06 : if (debuglvl > 0)
                        fprintf(debuglog, "rtc_r: port=0x%02X", port);
//                     printf("rtc_r: port=0x%02X", port);
                     return 0;
         case 0x07 : data = rtc.ram[addr];
                     if (addr < reg_a)
                        {
                         if (((addr == hours) || (addr == hours_alarm)) && ((rtc.member.reg_b & RTC_B_2412) == 0))
                            {
                             if (data == 0)
                                data = 12;
                             else
                                if (data > 12)
                                   data %= 12;
                            }       
                         if ((rtc.member.reg_b & RTC_B_DM) == 0)
                            data = ((data / 10) << 4) | (data % 10);
                        }
                     else
                        if (addr == reg_c)
                           rtc.member.reg_c = 0;	// clear all flag bits
                           
                     if (debuglvl > 0)
                        fprintf(debuglog, "rtc_r: port=0x%02X data=0x%02X\n", port, data);
//                     printf("rtc_r: port=0x%02X data=0x%02X\n", port, data);
                     return data;
        }            
    }
 return 0;
}

//==============================================================================
// RTC write register data - Port function
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void rtc_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (rtc_use)
    {
#if 0
     if ((standard != 0) && (port > 0x07))
        return 0;
#endif
     switch (port & 0x0f)
        {
         case 0x04 : addr = (data & 0x3F);
                     break;
         case 0x06 : if (addr < reg_a)
                        rtc.ram[addr] = data;
                     else
                        if (addr == reg_b)
                           {
                            if ((data & RTC_B_SET) && ((rtc.ram[addr] & RTC_B_SET) == 0))
                               data &= (0xff ^ RTC_B_UIE);
                            rtc.ram[addr] = data;
                           }
                        else
                           if (addr > reg_d)
                              rtc.ram[addr] = data;
                     break;
         case 0x07 : break;
        }             
     if (debuglvl > 0)
        fprintf(debuglog, "rtc_w: port=0x%02X data=0x%02X\n", port, data);
//     printf("rtc_w: port=0x%02X data=0x%02X\n", port, data);
    }
}

//==============================================================================
// RTC poll
//
// Read the host system time and check for any RTC interrupts events that may
// need to be generated.
//
// Only the Periodic (set per second for now) and the Update ended interrupts
// are currently supported.
//
//   pass: void
// return: int				RTC interrupt status
//==============================================================================
int rtc_poll (void)
{
 static int secs1 = -1;
 int secs;

 if (rtc_use)
    {
     rtc_setclockfromhost();

     secs = (z80_cycles / 3375000);
     if (secs != secs1) 
        {
         secs1 = secs;

         rtc.member.reg_c |= RTC_C_PF;		// set periodic flag
//         rtc.member.reg_c |= RTC_C_AF;        // set alarm  flag         
         rtc.member.reg_c |= RTC_C_UF;		// set update ended flag

         if ((rtc.member.reg_b & rtc.member.reg_c) & (RTC_B_PIE | RTC_B_AIE | RTC_B_UIE))
            rtc.member.reg_c |= RTC_C_IRQF;	// set IRQF flag
        }
     if (rtc.member.reg_c & RTC_C_IRQF)
        return B8(10000000);
    }
    
 return 0;
}
