//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                              uBee512 module                                *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Provides init, application loop, and deinit.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.0.0 - 13 October 2007, S.J.Kay
// - Confirmation windows added for Unices builds.
// - Added --nodisk option.
// - Added --rtc option and rtc related functions.
// - Rearranged ordering from most recent to oldest for FDD models and enabled
//   the PJB P256K, 256K and 256TC models.
// - APC model name changed to 56K.
//
// v1.4.0 - 6 October 2007, S.J.Kay
// - Added --mmode option to force the FDD ROM's monitor mode to start.
// - Added --pcg option to select amount of PCG RAM for emulation (Premium).
// - Added --vdu option to select amount of VDU RAM for emulation (Premium).
// - Added --aspect option to select 1:1 or 2:1 Y scaling.
// - Changes for console input/output for the windows port, adding -mwindows
//   to the linking stopped the background console prompt appearing.
// - Added (Windows) --conio option to allow console to be output at all
//   times, else only outputs for --help and --version options or when an
//   error causes the emulator to exit.  If any console output was made a
//   message box appears requiring confirmation to allow the console output
//   to be read before exiting.  By default all stdout and stderr output
//   will redirected to files stdout.txt and stderr.txt files (see SDL).
// - Close window mouse click on [X] is now functional with a confirmation
//   request before exiting in Windows build.
// - Confirmation request for exit and reset keys added for Windows build,
//   ESC+RESET will not request a confirmation.
// - Added --model option.  This allows other model Microbees to be emulated.
// - Changes for win32 build to allow installation to any directory.  This now
//   allows an installer to be used instead of a ZIP file.
// - Main event handler moved out of keyb.c module an used here to dispatch to
//   each event handler as needed. 
// - string_search() function now returns matching index value and -1 if no
//   match found.
// - option --sound=off now does not init sound in SDL. 
// - Start some code re-organisation.
//
// v1.3.0 - 1 September 2007, S.J.Kay
// - Added PIO polling and PIO configure (for interrupts) with execution
//   ratio change feature when interrupts are made active.
// - Added RS232 serial TX and RX options to set communication port,  baud
//   rates for each direction, data and stop bits: --coms, --baud, --baudtx,
//   --baudrx, --datab and --stopb.
// - Added printer output options --print and --printa.
// - Additional directory 'printer' created in the user's home account in the
//   .ubee512 directory.
// - Added --tapevol option for setting volume level of tape wave file.
// - Changed some error messages to put out function name first.
//
// v1.2.0 - 21 August 2007, S.J.Kay
// - Added --tapeo, --tapei, and --tapesamp options to support tape functions.
// - Additional directory 'tapes' created in the user's home account in the
//   .ubee512 directory.
// - Added function module calls for init, deinit and reset.
// - disk_and_roms() function name changed to set_account_paths()
//
// v1.1.0 - 18 August 2007, S.J.Kay
//   Code has been modified to allow the XTAL frequency to be specified.  In
//   theory if the host PC can emulate the Z80 code at that speed then that
//   is what you will get with a frame rate of (n) FPS.
//
//   See the README file for an explanation of how the speed emulation works.
//
// - stdout and stderr on windows now directed back to the console instead
//   of stdout.txt and stderr.txt files.
// - Modified the cursor blinking and Premium hardware inverse/flashing
//   video timer methods depending on if turbo mode is used.  This keeps the
//   flash rate correct.
// - Added frate command line option, mainly intended for hacking.
// - Added vblank command line option, mainly intended for use with the
//   turbo option to make the host timer determine the 50Hz VBLANK rate.
// - Added XTAL command line option (-x) where the xtal rate in MHz can be
//   specified in floating point format.
// - Added sound command line options: off, prop and normal.
// - Added volume setting command line option.
// - Changed sound hold-off to use a timer instead of Z80 cycle counter.
// - Changed from getopt to getopt_long command arguments. Uses same single
//   options as before.
// - Drive B not responding if a drive D was specified.  Problem caused by
//   floppyd[0] incorrectly labeled as floppyb[0] in the disk_and_roms()
//   function clearing any values already set for floppyb.
//
// v1.0.1 - 11 August 2007, S.J.Kay
// - Wrong default disk image type set for boot. Was boot.dip,  changed to
//   boot.dsk
//
// v1.0.0 - 8 August 2007, S.J.Kay
// - A major recoding of most of the original source to improve speed
//   regulation and for the addition of new features.
// - Create directories in users home accounts for the placement of disk and
//   ROM images for first time use.
// - Z80_cycles and associated variables now use unsigned 64 bit variables.
// - Added signal call back handler to allow control from other applications.
// - Added video timers and function calls in application loop for blinking
//   cursor and alpha+ hardware video flashing.
// - Added some code in the application loop to support the new sound emulation
//   feature.
// - Recoded the Z80 CPU speed regulation for the clock target.  The code now
//   uses the gettimeofday() function for unices and clock() function in mingw
//   builds to determine delay periods between mz80exec(n) function calls.
// - Added command line option for standard and standard+colour emulation.
// - Added command line monitor type options.
// - Added command line parameters for disk images for drives A-D.
// - Added turbo command line option.
// - Added fullscreen command line option.
// - Report init failures, not successes.
//
// v0.0.0 - 16 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//============================================================================== 
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================   

#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <time.h>
#include <SDL.h>

#ifdef MINGW
#include <windows.h>
#include <conio.h>
#else
#include <sys/types.h> 		// various type definitions, like pid_t
#include <sys/time.h>
#include <sys/stat.h>
#include <signal.h>		// signal name macros, and the signal() prototype
#endif

#include "ubee512.h"
#include "z80.h"
#include "vdu.h"
#include "crtc.h"
#include "memmap.h"
#include "keyb.h"
#include "fdc.h"
#include "mz80.h"
#include "gui.h"
#include "paknet.h"
#include "pio.h"
#include "sound.h"
#include "tape.h"
#include "serial.h"
#include "rtc.h"
#include "function.h"

//==============================================================================
// structures and variables
//==============================================================================
FILE *debuglog;
int  debuglvl;

uint64_t z80_cycles;

int key_esc;
int key_quit;
int key_reset;

int fullscreen;
int standard;

int flashvideo;

int done;
int loop;

#ifdef MINGW
char execpath[512];
clock_t ticks1, ticks2;
int consoleoutput;
#else
static struct timeval tod;
uint64_t ticks_tod;

char *shared_images[] = {"ubee512-logo.bmp", NULL};

#endif

int delay;

char userfile[512];
char destfile[512];

char userhome[512];
char userhome_docspath[512];
char userhome_diskpath[512];
char userhome_romspath[512];
char userhome_tapepath[512];
char userhome_prnapath[512];
char userhome_prnbpath[512];
char userhome_imagepath[512];
char userhome_rtcpath[512];

char floppya[512];
char floppyb[512];
char floppyc[512];
char floppyd[512];

char tapei[512];
char tapeo[512];

char printa[512];
char printb[512];

char coms1dev[512];

#ifdef use_audio
clock_t ticksound;
#endif

SDL_Event event;

volatile char gui_signal;

int sound = 1;			// sound on by default
static int turbo;

char *floppy_a = NULL;
char *floppy_b = NULL;
char *floppy_c = NULL;
char *floppy_d = NULL;

char *tape_i = NULL;
char *tape_o = NULL;
static int tape_samples=TAPE_SAMPLE_FREQ;
static int tape_volume=TAPE_VOLUME;

char *print_a = NULL;
char *print_b = NULL;

char *coms1_dev = NULL;
int serial_tx_baud = SERIAL_TX_BAUD;
int serial_rx_baud = SERIAL_RX_BAUD;
int serial_databits = SERIAL_DATABITS;
int serial_stopbits = SERIAL_STOPBITS;

static float opt_xtal = XTALFREQ; // default XTAL frequency 
int frame_rate = FRAMERATE;	// default frame rate
extern int audio_volume;
int vblank_method = 0;

int model=1;
int rom_model=0;
int monitor=0;			// 0=colour monitor
int colour_use=1;		// 0=no colour, 1=use colour 
int rtc_use=1;
int ram=512;
int boot_addr=0x8000;		// start executing code here
int xtal;
int z80ms;

int z80_ratio;			// current z80 execution ratio (default = 1)

char *model_args[] =
{
 "256tc",
 "p512k",
 "512k",
 "p128k",
 "128k",
 "p256k",
 "256k",
 "p64k",  
 "64k",
 "56k",
 "ppc85",
 "pc85",  
 "pc",
 "ic",
 "2mhz"
 };

static int opt_aspect=2;
static int opt_pcg=32;
static int opt_vdu=2;

static int z80_block_cycles_def; // default number of Z80 cycles in 1 block
static int z80_blocks_def;	 // default number of Z80 blocks

static int z80_block_cycles_cur; // current number of Z80 cycles in 1 block
static int z80_blocks_cur;	// current number of Z80 blocks

int z80_block_cycles;		// working number of Z80 cycles in 1 block
int z80_blocks;			// working number of Z80 blocks

extern int fdc_status0;

extern int extendram;
extern int crtcresized;

extern int cur_blink;
extern int cur_mode;
extern int cur_pos;

extern int audio_silence;
extern int dobuffering;

//==============================================================================
// Search an array of strings for the first occurence of the passed search
// string.  The string array must be be terminated by an empty string. 
//
//   pass: char *strg_array	pointer to a string array to be checked
//         char *strg_find	pointer to a string to search for
// return: int			index if found, else -1
//==============================================================================
int string_search (char *strg_array[], char *strg_find)
{
 int i;
 
 for (i=0; strg_array[i][0] != 0; i++)
    {
     if (strcmp(strg_find, strg_array[i]) == 0)
        return i;
    }
 return -1;
}        

#ifdef MINGW
#else
//==============================================================================
// External GUI signal handler.
//
// This allows an optional external GUI program to change values during the
// running of the emulator.  The code here is kept as simple as possible and
// simply just sets a flag to indicate that a signal was received.  The main
// loop checks the flag value to see if any action is required.
//
// Can be tested in unices by using the following to cause a reset():
// $kill -s 30 pid
//
// To find the pid use:
// $ps ax | grep ubee512
//
//   pass: int sig_num
// return: void
//==============================================================================
void signal_handler (int sig_num)
{
 gui_signal = 1;
}
#endif

//==============================================================================
// Event checking.
//
//   pass: void
// return: void
//==============================================================================
void event_handler (void)
{
 while (SDL_PollEvent(&event))
    {
     switch (event.type)
        {
         case SDL_KEYDOWN : keyb_keydown_event();
                            break; 
         case   SDL_KEYUP : keyb_keyup_event();
                            break;
         case    SDL_QUIT : 
                            if (gui_message_box(BUTTON_OKCANCEL,
                            "All unsaved data will be lost!\n\n" \
                            "Exit uBee512 ?") == BUTTON_IDOK)
                               done = 1;
                            break;
        }                    
    }
}

//==============================================================================
// Create user home account directories and determine paths to be for disks,
// ROMs, tapes, and printers.
//
//   pass: void
// return: int				0 if no errors, else -1
//==============================================================================
int set_account_paths (void)
{
 int i;

#if MINGW
#define SLASHCHAR '\\'
#define DIR_DOCS "\\doc\\"
#define DIR_DISKS "\\disks\\"
#define DIR_ROMS "\\roms\\"
#define DIR_TAPES "\\tapes\\"
#define DIR_PRINTER "\\printer\\"
#define DIR_IMAGES "\\images\\"
#define DIR_RTC "\\rtc\\"

// TEMPORARY - This will be changed in a later version
#define execpath userhome

// get the path and executable filename
 if (GetModuleFileName(NULL, execpath, sizeof(execpath)) == 0)
    {
     console_output();    
     printf("ubee512: Unable to find the executable path\n");
     return -1;
    }

// delete the executable name part and last '\' character
 i = strlen(execpath);
 while (execpath[--i] != SLASHCHAR)
    {}
 execpath[i] = 0;

#if 0
// get the environment path for application data
 if (GetEnvironmentVariable("APPDATA", userhome, sizeof(userhome)) == 0)
    {
     console_output();    
     printf("ubee512: Unable to obtain the APPDATA environment\n");
     return -1;
    }
 strcat(userhome, "/ubee512"); 

// printf("env var=%s\n", userhome);
#endif

#else

#define SLASHCHAR '/'
#define DIR_DOCS "/doc/"
#define DIR_DISKS "/disks/"
#define DIR_ROMS "/roms/"
#define DIR_TAPES "/tapes/"
#define DIR_PRINTER "/printer/"
#define DIR_IMAGES "/images/"
#define DIR_RTC "/rtc/"

 char *s;
 int dirres;
 mode_t mode;
 FILE *textfp;
 
 s = getenv("HOME");
 if (s == NULL)
    {
     console_output();
     printf("ubee512: Unable to find the user home path\n");
     return -1;
    }
 else
    strcat(userhome, s);

 strcat(userhome, "/.ubee512"); 
#endif
    
 strcat(userhome_docspath, userhome); 
 strcat(userhome_docspath, DIR_DOCS);

 strcat(userhome_diskpath, userhome); 
 strcat(userhome_diskpath, DIR_DISKS);

 strcat(userhome_romspath, userhome);
 strcat(userhome_romspath, DIR_ROMS);

 strcat(userhome_tapepath, userhome);
 strcat(userhome_tapepath, DIR_TAPES);

 strcat(userhome_prnapath, userhome);
 strcat(userhome_prnapath, DIR_PRINTER);

 strcat(userhome_prnbpath, userhome);
 strcat(userhome_prnbpath, DIR_PRINTER);

 strcat(userhome_imagepath, userhome);
 strcat(userhome_imagepath, DIR_IMAGES);
 
 strcat(userhome_rtcpath, userhome);
 strcat(userhome_rtcpath, DIR_RTC);

#ifdef MINGW
#else
 mode = S_IRWXG | S_IRWXU | S_IROTH | S_IXOTH;
 
 dirres = mkdir(userhome, mode);
 if (dirres == 0)	// if there was no directory and it was created ok
    {
// create the directories in the user's home account
     dirres = mkdir(userhome_docspath, mode);
     dirres = mkdir(userhome_diskpath, mode);
     dirres = mkdir(userhome_romspath, mode);
     dirres = mkdir(userhome_tapepath, mode);
     dirres = mkdir(userhome_prnapath, mode);
     dirres = mkdir(userhome_prnbpath, mode);
     dirres = mkdir(userhome_imagepath, mode);
     dirres = mkdir(userhome_rtcpath, mode);     

// copy files from the shared directory to the user's home account
     for (i = 0; shared_images[i] != NULL; i++)
        {
         snprintf(userfile, sizeof(userfile), PATH_SHARED_IMAGES"%s", shared_images[i]);
         snprintf(destfile, sizeof(destfile), "%s%s", userhome_imagepath, shared_images[i]);
         copy_file(destfile, userfile);
        } 

// create readme.txt file telling the user where to find documentation
     snprintf(userfile, sizeof(userfile), "%sreadme.txt", userhome_docspath);
     textfp = fopen(userfile, "w");
     fprintf(textfp, "Documentation including licensing information is located in:\n");
     fprintf(textfp, PATH_SHARED_DOCS"\n");
     fclose(textfp); 

     console_output();
     printf("\n");
     printf("=========================================================================\n");
     printf("This is the first time you have run the ubee512 emulator in this account.\n");
     printf("Directories have been created in your home account for first time use.\n");
     printf("\n");
     printf("You now need to copy ROMs: charrom.bin, and either disk and/or BASIC ROMs\n");
     printf("to the ROMs directory:\n");
     printf("%s\n", userhome_romspath);
     printf("\n");
     printf("Also copy any required disk image(s) to:\n");
     printf("%s\n", userhome_diskpath);     
     printf("\n");     
     printf("When you have done this run the ubee512 emulator again.\n");          
     printf("=========================================================================\n");
     exit(0);		// exit so user can copy the required files
    } 
#endif

 if (! rom_model)
    {    
// Image A image
     if (floppy_a != NULL)
        {
         if (strchr(floppy_a, SLASHCHAR) == NULL)
            {
             floppya[0] = 0;
             strcat(floppya, userhome_diskpath);
             strcat(floppya, floppy_a); 
            }
         else    
            strcat(floppya, floppy_a);
        }

// Image B image
     if (floppy_b != NULL)
        {
         if (strchr(floppy_b, SLASHCHAR) == NULL)
            {
             floppyb[0] = 0;
             strcat(floppyb, userhome_diskpath);      
             strcat(floppyb, floppy_b); 
            }
         else    
            strcat(floppyb, floppy_b);
        }

// Image C image
     if (floppy_c != NULL)
        {
         if (strchr(floppy_c, SLASHCHAR) == NULL)
            {
             floppyc[0] = 0;
             strcat(floppyc, userhome_diskpath);      
             strcat(floppyc, floppy_c); 
            }
         else    
            strcat(floppyc, floppy_c);
        }

// Image D image	
     if (floppy_d != NULL)
        {
         if (strchr(floppy_d, SLASHCHAR) == NULL)
            {
             floppyd[0] = 0;
             strcat(floppyd, userhome_diskpath);      
             strcat(floppyd, floppy_d); 
            }
         else    
            strcat(floppyd, floppy_d);
        }
    }

// Tape input file
 if (tape_i != NULL)
    {
     if (strchr(tape_i, SLASHCHAR) == NULL)
        {
         tapei[0] = 0;
         strcat(tapei, userhome_tapepath);      
         strcat(tapei, tape_i); 
        }
     else    
        strcat(tapei, tape_i);
     printf("Tape in file: %s\n", tapei);
    }

// Tape output file
 if (tape_o != NULL)
    {
     if (strchr(tape_o, SLASHCHAR) == NULL)
        {
         tapeo[0] = 0;
         strcat(tapeo, userhome_tapepath);      
         strcat(tapeo, tape_o); 
        }
     else    
        strcat(tapeo, tape_o);
     printf("Tape out file: %s\n", tapeo);
    }

 if ((tape_o != NULL) && (tape_i != NULL))
    if (strcmp(tapeo, tapei) == 0)
       {
        console_output();
        printf("ubee512: tape in and out can not be the same file\n");
        return -1;
       }

// Printer (ASCII) output file
 if (print_a != NULL)
    {
     if (strchr(print_a, SLASHCHAR) == NULL)
        {
         printa[0] = 0;
         strcat(printa, userhome_prnapath);      
         strcat(printa, print_a); 
        }
     else    
        strcat(printa, print_a);
     printf("Printer (ASCII) file: %s\n", printa);
    }

// Printer (BINARY) output file
 if (print_b != NULL)
    {
     if (strchr(print_b, SLASHCHAR) == NULL)
        {
         printb[0] = 0;
         strcat(printb, userhome_prnapath);      
         strcat(printb, print_b); 
        }
     else    
        strcat(printb, print_b);
     printf("Printer (BINARY) file: %s\n", printb);
    }

// Serial communications device #1
 if (coms1_dev != NULL)
    {
     strcat(coms1dev, coms1_dev);
     printf("Serial communications dev: %s\n", coms1dev);
    }

 return 0;
}

//==============================================================================
// Initialise.
//
//   pass: void
// return: int				0 if no errors, else -1
//==============================================================================
static int init (void)
{
 SDL_Surface *icon;
 uint32_t sdl_init_properties;

 printf(TITLESTRING"\n");

 if (set_account_paths() != 0)
    return -1;
 
#ifdef MINGW
#else
// set the GUI signal handler
 signal(30, signal_handler);
#endif

 if (debuglvl > 0)
    {
     debuglog = fopen(LOGFILE, "w"); 
     if (debuglog == NULL)
        {
         console_output();
         printf("init: Unable to write to log file\n");
         return -1;
        }
     fprintf(debuglog, "debug level = %i\n", debuglvl);
    }

 sdl_init_properties = SDL_INIT_VIDEO | SDL_INIT_TIMER;
 if (sound)
    sdl_init_properties |= SDL_INIT_AUDIO;

 if (SDL_Init(sdl_init_properties) != 0)
    {
     console_output();
     printf("init: Failed SDL_Init - %s\n", SDL_GetError());
     return -1;
    }

 SDL_EventState(SDL_ACTIVEEVENT | SDL_KEYUP | SDL_KEYDOWN | SDL_SYSWMEVENT, SDL_IGNORE);

 SDL_WM_SetCaption(TITLESTRING, ICONSTRING);

 userfile[0] = 0;
 strcat(userfile, userhome_imagepath);
 strcat(userfile, "ubee512-logo.bmp"); 

 if ((icon = SDL_LoadBMP(userfile)) != NULL)
    {
     SDL_SetColorKey(icon, SDL_SRCCOLORKEY, SDL_MapRGB(icon->format, 0, 0, 0));
     SDL_WM_SetIcon(icon, NULL);
     SDL_FreeSurface(icon);
    }

 if (gui_init() != 0)
    {
     console_output();
     printf("init: Failed gui_init\n");    
     return -1;
    } 
 else
    {
    }   

 if (z80_init() != 0)
    {
     console_output();
     printf("init: Failed z80_init\n");    
     return -1;
    } 

 if (vdu_init() != 0)
    {
     console_output();
     printf("init: Failed vdu_init\n");
     return -1;
    } 

 if (memmap_init() != 0)
    {
     console_output();
     printf("init: Failed memmap_init\n");    
     return -1;
    } 

 if (crtc_init() != 0)
    {
     console_output();
     printf("init: Failed crtc_init\n");
     return -1;
    }

 if (keyb_init() != 0)
    {
     console_output();
     printf("init: Failed keyb_init\n");
     return -1;
    } 

 if (fdc_init() != 0)
    {
     console_output();
     printf("init: Failed fdc_init\n");
     return -1;
    } 

 if (pio_init() != 0)
    {
     console_output();
     printf("init: Failed pio_init\n");
     return -1;
    } 

 if (paknet_init() != 0)
    {
     console_output();
     printf("init: Failed paknet_init\n");
     return -1;
    } 

 if (rtc_init() != 0)
    {
     console_output();
     printf("init: Failed rtc_init\n");
     return -1;
    } 

 if (function_init() != 0)
    {
     console_output();
     printf("init: Failed function_init\n");
     return -1;
    } 

 return 0;
}

//==============================================================================
// De-initialise.
//
//   pass: void
// return: int			result of de-init functions
//==============================================================================
static int deinit(void)
{
 int res;

 if (debuglvl > 0)
    fclose(debuglog);

 res = function_deinit() || pio_deinit() || rtc_deinit() || paknet_deinit() ||
       fdc_deinit() || vdu_deinit() || keyb_deinit() || crtc_deinit() ||
       memmap_deinit() || z80_deinit() || gui_deinit();

 SDL_Quit();

 return res;
}

//==============================================================================
// Reset the virtual Microbee.
//
// Make sure all reset functions clear variables that are constantly active
// values.
//
//   pass: void
// return: int			result of reset functions
//==============================================================================
static int reset (void)
{
 int res;
 
 z80_cycles = 0;
 done = 0;
 loop = 0;

#ifdef use_resetinfo
 printf("ubee512: emulation reset\n");
#endif

#ifdef MINGW
 ticks1 = clock();
#else
 gettimeofday(&tod, NULL);          
 ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif
 
 res =  z80_reset() || memmap_reset() || crtc_reset() || keyb_reset() ||
        vdu_reset() || fdc_reset() || pio_reset() || paknet_reset() ||
        rtc_reset() || function_reset();

 return res;
}

//==============================================================================
// Z80 execution ratio set.
//
// The z80 execute method may be changed to execute fewer Z80 cycles in one Z80
// block but with more blocks by changing the ratio of these two values. 
//
// This is intended to allow PIO interrupts to be serviced faster as each
// interrupt is only serviced once on each call to mz80exec().  The ratio
// used can be set back to default after interrupts have completed.
//
//   pass: int divider		use 1 for default values
// return: void
//==============================================================================
void z80_exec_ratio (int divider)
{
 if (divider == 1)
    {
     z80_block_cycles_cur = z80_block_cycles_def;
     z80_blocks_cur = z80_blocks_def;
    }
 else
    {
     z80_block_cycles_cur = z80_block_cycles_def / divider;
     z80_blocks_cur = divider;
    } 

 z80_ratio = divider;
}

//==============================================================================
// Model configuration
//
//   pass: void
// return: int			0 if model supported, else -1
//==============================================================================
int model_config (void)
{
 colour_use = 1;
 
 switch (model)
    {
     case MOD_256TC : ram = 256;		// 256TC DRAM FDD model
                      break;
     case MOD_P512K : ram = 512;		// PJB upgraded Premium 128K DRAM FDD model
                      break;
     case  MOD_512K : ram = 512;		// PJB upgraded 128K DRAM FDD model
                      standard = 1;
                      break;
     case MOD_P256K : ram = 256;		// PJB upgraded Premium 64K DRAM FDD model
                      break;
     case  MOD_256K : ram = 256;		// PJB upgraded 64K DRAM FDD model
                      standard = 1;
                      break;
     case MOD_P128K : ram = 128;		// Premium 128K (SBC) DRAM FDD model
                      rtc_use = 0;
                      break;
     case  MOD_128K : ram = 128;		// 128K (SBC) DRAM FDD model
                      rtc_use = 0;
                      standard = 1;
                      break;
     case  MOD_P64K : ram = 64;			// Premium 64K DRAM FDD model
                      rtc_use = 0;
                      console_output();
                      break;
     case   MOD_64K : ram = 64;			// CIAB 64K DRAM FDD model
                      rtc_use = 0;
                      standard = 1;
                      console_output();
                      break;
     case   MOD_56K : ram = 56;
                      rtc_use = 0; 
                      standard = 1;
                      colour_use = 0;		// use mono, used old colour method
                      boot_addr=0xe000;
                      console_output();
                      break;
     case MOD_PPC85 : rom_model = 1;		// Premium PC85
                      ram = 32;
                      rtc_use = 0;
                      break;
     case  MOD_PC85 : rom_model = 1;		// Standard PC85
                      standard = 1;     
                      ram = 32;
                      rtc_use = 0;
                      break;
     case    MOD_PC : rom_model = 1;		// Personal Communicator
                      standard = 1;
                      ram = 32;
                      rtc_use = 0;
                      break;
     case    MOD_IC : rom_model = 1;		// IC means ???
                      standard = 1;
                      colour_use = 0;		// no colour
                      ram = 32;
                      rtc_use = 0;
                      break;
     case  MOD_2MHZ : opt_xtal = 2.0;		// Microbee speed
                      rom_model = 1;
                      standard = 1;
                      colour_use = 0;		// no colour
                      ram = 32;
                      rtc_use = 0;
                      break;
    }
 return 0;
}

//==============================================================================
// Application setup.
//
//   pass: void
// return: void
//==============================================================================
void ApplicationSetup (void)
{
 float speedadj;
 float z80ms_r;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// XTAL configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// convert XTAL to an integer in Hz
 xtal = (int)(opt_xtal * 1E6);

// calculate the number of Z80 cycles executed in 1 frame 
 z80_block_cycles_def = xtal / frame_rate;

// execution time in mS needed for one lot of z80_block_cycles_def
 z80ms_r = ((float)z80_block_cycles_def / xtal) * 1000.0;

// calculate a speed adjustment based on a constant percentage
 speedadj = (float)z80ms * (SPEED_PCT_ADJ / 100.0);

// apply the adjustment value
 z80ms = (int)(z80ms_r - speedadj);
 if (z80ms < 0)
    z80ms = 0;

 z80_blocks_def = 1;	// default blocks

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Z80 PIO configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 pio_configure(xtal, frame_rate, serial_rx_baud); // set exec ratios prota/serial
 
 z80_exec_ratio(1);	// use the values calculated here

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Tape configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 tape_config_out(xtal, tape_samples, tape_volume);
 tape_config_in(xtal);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// RS232 Serial configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 serial_config(xtal, serial_tx_baud, serial_rx_baud, serial_databits, serial_stopbits);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Sound configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 switch (sound)
    {
     case 0 : break;
     case 1 : sound_clock(xtal, SOUND_PROP); // sound proportional to XTAL freq
              break;
     case 2 : sound_clock(xtal, SOUND_3375); // sound set to normal rate
              break; 
    }             

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// CRTC configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 crtc_configure(opt_aspect);
 crtc_clock(xtal, vblank_method);  // set VBLANK method

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// VDU configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 vdu_configure(opt_pcg, opt_vdu);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// reset and preset timers
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 reset();			// resets the virtual Microbee

#ifdef MINGW
 ticks1 = clock();   
#else
 uint64_t ticks_tod;
 gettimeofday(&tod, NULL); 
 ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif 
}

//==============================================================================
// Application loop.
//
// Video timers
// ------------
// The CRTC blinking cursor and Alpha+ hardware inverse and flashing video
// timers use the accumulated Z80 cycle count,  this works well as the Z80 CPU
// execution is regulated with delays and so the instruction count is accurate
// over a period of time,  the rate will increase in turbo mode.
//
// Invalid instruction
// -------------------
// There appears to be a bug in v3.4 of makez80 or an uninitialised area of
// memory in another module.  It consistently occurs at 0xdec1 after
// 12,210,520 cycles are executed and a reset is then needed to fix the
// situation. The error occurs regardless of what CP/M OS is booted, the
// error still occurs when filling the boot track with 0s and a "jr $". Have
// found this to be a problem in win32 build but may be present on other
// builds too. There may be a variable not initilised correctly and only
// showing up in the win32 build. an old v2.7 of this code appears OK in
// this area but has other problems in itself.  The work around for now is
// to just reset and restart on each invalid instruction encountered.  This
// is NOT the end solution as some un-intened z80 instructions could get
// executed and may not even get an invalid instruction. It was found that
// the minimum work around required the following but decided to do a full
// reset instead. The minimum work around should give some clues.  A bank
// select issue could be causing rubbish to be exectued.
//
// z80_cycles = 0;   
// z80_reset();
// memmap_reset();
//
//   pass: void
// return: void
//==============================================================================
void ApplicationLoop (void)
{
 int dwResult;
 struct mz80context z80; 

 int cur_blink_rate;
 int cur_blink_last=0;
 int flashvideo_last=0;
 int invalid = 0;
 
 while (! done)
    {
#ifdef MINGW
     clock_t ticks_s; 		// for sound hold-off period
     ticks1 = clock();
#else
     uint64_t ticks_tod_snd;    // for sound hold-off period
     struct timeval tod_snd;
     struct timeval tod_x;     
     gettimeofday(&tod, NULL);          
     ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif

// looping feature for speeding up PIO interrupt responses by having a
// smaller z80_block_cycles value.  Normally z80_blocks is set to 1 with
// z80_block_cycles equal to the default calculated rate.

     z80_blocks = z80_blocks_cur;
     z80_block_cycles = z80_block_cycles_cur;
   
     while (z80_blocks--)
        {
// Execute a block (Z80CYCLES) of Z80 instructions and return the result.
// The Z80CYCLES value used has been calculated for timing purposes.
         dwResult = mz80exec(z80_block_cycles);

// If the result is not 0x80000000, it's an address where an invalid
// instruction was hit.
         if (dwResult != 0x80000000)
            {
             mz80GetContext(&z80);
             if (++invalid > 1)
                {
                 console_output();
                 printf("Invalid instruction at: 0x%4x\n", (unsigned int)z80.z80pc);
                 printf("           Clock ticks: %ld\n", z80.z80clockticks);
                } 
             z80_cycles = 0;   
             reset();
            }

         z80_cycles += mz80GetElapsedTicks(1);

// poll the PIO for interrupt events
         pio_polling();
        } 

// Resize the video when certain CRTC registers have been changed under Z80
// emulation. A wait period is introduced here to prevent the display being
// resized too often in a short time period.
     if (crtcresized)
        {
         if (crtcresized++ == 2)	// will be the 2nd time around loop
            crtc_videochange();		// resets crtcresized
        }

// Determine the current status for the CRTC blinking cursor and refresh it
// if this has changed.  The method used here depends on if turbo mode is
// used. if turbo mode is used then Z80 execution speed will not be known as
// no delays will be inserted, if not turbo then the rate must be determined
// by the Z80 cycle count to achieve smooth results.
     if (cur_mode > 1)			// only if blinking cursor
        {
         if (turbo)
            {
             if (cur_mode == 2)
                cur_blink_rate = 500;	// blinking at 1/32 field rate
             else   
                cur_blink_rate = 250;	// blinking at 1/16 field rate
#ifdef MINGW
             if ((clock() / cur_blink_rate) & 0x01)
#else
             gettimeofday(&tod_x, NULL);          
             if (((tod_x.tv_usec / 1000) / cur_blink_rate) & 0x01) // rolls over each sec
#endif
                cur_blink = 0xFF;
             else
                cur_blink = 0x00;
            }
         else
            {
             if (cur_mode == 2)
                cur_blink_rate = xtal / 2;	// blinking at 1/32 field rate
             else   
                cur_blink_rate = xtal / 4;	// blinking at 1/16 field rate        
             if ((z80_cycles / cur_blink_rate) & 0x01)
                cur_blink = 0xFF;
             else
                cur_blink = 0x00;
            }            
         if (cur_blink != cur_blink_last)
            {
             cur_blink_last = cur_blink;            
             crtc_redraw_char(cur_pos);
            }
        }

// Determine the current state of the alpha+ flashing video and refresh it
// if this has changed.  the method used here is the same as that used by
// the blinking cursor timer.
     if (extendram)			// only if extended RAM selected
        {
         if (turbo)
            {
#ifdef MINGW
             if ((clock() / 250) & 0x01)
#else
             gettimeofday(&tod_x, NULL);          
             if (((tod_x.tv_usec / 1000) / 250) & 0x01) // rolls over each sec
#endif
               flashvideo = 1;
            else
               flashvideo = 0;
            }
         else
            {
             if ((z80_cycles / (xtal / 4)) & 0x01)
                flashvideo = 1;
             else
                flashvideo = 0;
            }
         if (flashvideo != flashvideo_last)
            {
             flashvideo_last = flashvideo;            
             crtc_set_redraw();	// force crtc to redraw screen
            }
        }    

#if use_audio
// Sound buffering.  This is required as CPU emulation speed is affected by
// other processes occuring at the same time, we need to try to ensure the
// sound is slightly in front so that it does not matter if the Z80 lags
// behind slightly in creating the sound.  A 100mS delay has been found
// to be about right.  If the buffer delay is too long the lag will be
// noticable by the user.  If too short then break up, and crackling will be
// evident.
     if (dobuffering)
        {
         if (ticksound == -1)		// if just starting the buffering
#ifdef MINGW
            {
             ticksound = 0;
             ticks_s = clock();
            } 
#else
            { 
             ticksound = 0;            
             gettimeofday(&tod_snd, NULL);          
             ticks_tod_snd = (uint64_t)((tod_snd.tv_sec * 1000) + (tod_snd.tv_usec / 1000));
            }
#endif
         else
            {
#ifdef MINGW
             if ((clock() - ticks_s) >= SOUND_HOLD_OFF)
#else
             gettimeofday(&tod_snd, NULL);          
             if ((((uint64_t)((tod_snd.tv_sec * 1000) + (tod_snd.tv_usec / 1000))) - ticks_tod_snd) >= SOUND_HOLD_OFF)
#endif
                sound_unsilence();	// allow sound to play
            }
        }    
#endif

// Update the display (only if it's update flag is set)           
     gui_redraw();

// Check and handle any pending events.
     event_handler();

// Handle external GUI signals
     if (gui_signal)
        {
         reset();
         gui_signal = 0;
        }

// Check for exit and reset key flags and get confirmation
     if (key_quit)
        {
         key_quit = 0;
         if (gui_message_box(BUTTON_OKCANCEL,
         "All unsaved data will be lost!\n\n" \
         "Exit uBee512 ?") == BUTTON_IDOK)
            done = 1;
        } 
     if (key_reset)
        {
         key_reset = 0;
         if (! key_esc)		// if ESC+RESET then no prompting
            {
             if (gui_message_box(BUTTON_OKCANCEL,
             "All unsaved data will be lost!\n\n" \
             "Reset uBee512 ?") == BUTTON_IDOK)
                reset();
            }
         else
            reset();
         key_esc = 0;
        }

// Insert delays to achieve correct emulated Z80 CPU speed.  The Z80 speed
// may be made a little faster to help prevent audio break up (in ubee512.h).
//
// The MINGW build relies on a 1mS timer (1ms granularity) and the delay
// required is calculated from the last block (Z80CYCLES) of Z80
// instructions.
//
// The non MINGW builds (unices) uses the gettimeofday function and this is
// converted to a 64 bit variable with a 1mS resolution.
//
     if (turbo)
        SDL_Delay(0);
     else
        {
#ifdef MINGW
// Find the execution delay required for the last (1/frame_rate)
         ticks2 = clock() - ticks1;
         if (ticks2 >= 0)		// if clock() has not wrapped around
            {                   	// else use the last delay value
             delay = z80ms - ticks2;
            }
#else
         gettimeofday(&tod, NULL);
         delay = z80ms - ((uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000)) - ticks_tod);
#endif
         
// Execute the delay period.  If a delay is not required then a delay of 0
// is used to give up the CPU for other processes.
         if (delay < 0)
            SDL_Delay(0);
         else
            SDL_Delay(delay);
        }
    }    
}

//==============================================================================
// Application usage
//
//   pass: void
// return: void
//==============================================================================
void ApplicationUsage (void)
{
 char const usage[] =
TITLESTRING"\n"
"\n"
"Usage: ubee512 [options]\n"
"\n"
"-a, --image_a=file      Image file path & filename for floppy drive A\n"
"-b, --image_b=file      Image file path & filename for floppy drive B\n"
"-c, --image_c=file      Image file path & filename for floppy drive C\n"
"-d, --image_d=file      Image file path & filename for floppy drive D\n"
"\n"
"                        If you specify a path as part of the image name then\n"
"                        it will be accessed from that area.  If only an\n"
"                        image name is specified the emulator will look for\n"
"                        it here:\n"
"\n"
"                        Windows:\n"
"                        installed location\\ubee512\\disks\\\n"
"\n"
"                        Unices:\n"
"                        your home path/.ubee512/disks/\n"
"\n"
"                        Hint: if you want to access an image in the current\n"
"                        directory you need to force it using the following\n"
"                        method:\n"
"\n"
"                        Windows:\n"
"                        ubee512 -a .\\myimage.dsk\n"
"\n"
"                        Unices:\n"
"                        ubee512 -a ./myimage.dsk\n"
"\n"
"--aspect=n              Set the display aspect, n=1 is 1:1,  default aspect\n"
"                        2:1 (n=2),  n may be set to 1 or 2.  1:1 scaling\n"
"                        may be enforced for some CRTC6545 display sizes'.\n" 
"\n"
"--baud=rate             Set serial communications baud rate for both TX and\n"
"                        RX.  A value from 1 to 19200 is allowed.  Default\n"
"                        rate is 300 baud.  If Individual baud rates are\n"
"                        required for TX and RX then use the --baudtx and\n"
"                        --baudrx options instead.  This value must match the\n"
"                        Microbee serial application's value.\n"
"\n"
"--baudtx=rate           Set serial communications baud rate for TX only.  A\n"
"                        value from 1 to 19200 is allowed.  Default rate is\n"
"                        300 baud.  This value must match the Microbee serial\n"
"                        application's value.\n"
"\n"
"--baudrx=rate           Set serial communications baud rate for RX only.  A\n"
"                        value from 1 to 19200 is allowed.  Default rate is\n"
"                        300 baud. This value must match the Microbee serial\n"
"                        application's value.\n"
"\n"
"--coms=port             Serial communications port for emulation of RS232.\n"
"                        On Unices specify a device, on Windows specify the\n"
"                        com port number.  No serial communications will be\n"
"                        emulated if this option is not specified.\n"
"\n"
"--conio                 Switches on verbose console output for Windows port.\n"
"                        By default only fatal errors and some option's output\n"
"                        is sent to the console.\n"
"\n"
"--datab=bits            Set serial communications number of data bits.  A\n"
"                        value from 5 to 8 is allowed.  Default value is 8\n"
"                        data bits.  This value must match the Microbee\n"
"                        serial application's value.\n"
"\n"
"--frate=fps             Frame rate,  the default is 30 FPS,  an integer value\n"
"                        between 1 and 1,000,000 is allowed.  This option is\n"
"                        intended for 'hacking' and code development use.\n"
"\n"
"-f, --fullscreen        Start in full screen mode (use ALT+ENTER to toggle).\n"
"\n"
"-h, --help, --usage     Display help information on command line usage.\n"
"\n"
"--mmode                 Forces the return of the 'M' key once when the\n"
"                        emulator is started.  This is used for jumping\n"
"                        directly into the ROM's Monitor mode.\n"
"\n"
"--model=type            Model type, if this option is not specified the p512k\n"
"                        model is emulated.  The 'p' denotes a Premium\n"
"                        variation in a model.  The following models are\n"
"                        supported: 256tc, p512k, 512k, p256k, 256k, p128k,\n"
"                        128k, p64k, 64k, 56k, ppc85, pc85, pc, ic, and 2mhz.\n"
"\n"
"-m, --monitor=type      Monitor type, if this option is not specified a\n"
"                        colour monitor is the default. <type> may be one\n"
"                        of the following:\n"
"\n"
"                        a : amber monitor\n"
"                        g : green monitor\n"
"                        w : white monitor, white foreground on black\n"
"                            background\n"
"                        b : black monitor, black foreground on white\n"
"                            background\n"
"                        c : colour monitor\n"
"\n"
"                        Note: This option by itself does not force the\n"
"                        emulation into a standard model Microbee, it's use\n"
"                        simply determines what sort of monitor is connected\n"
"                        to the emulated Microbee, be it a standard or\n"
"                        Premium model.  If you want to force the Premium\n"
"                        graphics 'off' you must uses the -s<col> option.  If\n"
"                        <col> = m then this option can then further set\n"
"                        the desired monochrome colour.\n"
"\n"
"--nodisk                Force no disks initially,  used to start some boot\n"
"                        ROMs in menu mode.\n"
"\n"
"--pcg=n                 Premium model option that sets size of PCG RAM to be\n"
"                        emulated.  n may be 16 or 32.  Standard premium\n"
"                        models are 16K.  Default is 32K.\n"
"\n"
"--print=file            Printer output to file,  the output is not modified.\n"
"                        The location of the file follows the same rule as\n"
"                        used for the drive image options except default\n"
"                        directory is 'printer'.\n"
"\n"
"--printa=file           Printer output to file,  the output is converted to\n"
"                        ASCII decimal. The location of the file follows the\n"
"                        same rule as used for the drive image options except\n"
"                        default directory is 'printer'.\n"
"\n"
"--rtc=n                 Real Time Clock (RTC) emulation. n=1 to enable, n=0\n"
"                        to disable.  The following models use RTC by\n"
"                        default: 256tc, p512k, 512k, p256k, 256k.\n"
"\n"
"--sound=method          Determine the method used for sound,  the default\n"
"                        method is 'prop':\n"
"\n"
"                           off : sound is turned off\n"
"                          prop : sound is proportional to XTAL frequency\n"
"                        normal : sound rate forced as if 3.375 MHz XTAL\n"
"\n"
"-s, --standard=col      Standard graphics Microbee emulation,  turn off\n"
"                        the Premium (alpha+) graphics and use a standard\n"
"                        green monitor. The <col> specifier allows for\n"
"                        standard monochrome or colour support:\n"
"\n"
"                        c : standard Microbee with colour\n"
"                        m : standard Microbee with Green monochrome\n"
"\n"
"                        If using this option with the 'm' parameter then the\n"
"                        -m<mode> option can also be specified to select the\n"
"                        monochrome colour that is preferred.\n"
"\n"
"--stopb=bits            Set serial communications number of stop bits.  A\n"
"                        value from 1 to 2 is allowed.  Default value is 1\n"
"                        data bits.  This value must match the Microbee\n"
"                        serial application's value for TX.\n"
"\n"
"--tapei=file            Tape input from a WAV file.  The location of the file\n"
"                        follows the same rule as used for the drive image\n"
"                        options except default directory is 'tapes'.\n"
"\n"
"--tapeo=file            Tape output to a WAV file.  The location of the file\n"
"                        follows the same rule as used for the drive image\n"
"                        options except default directory is 'tapes'.\n"
"\n"
"--tapesamp=frequency    Tape output sample frequency in Hz.  Default is\n"
"                        22050 Hz.\n"
"\n"
"--tapevol=level         Tape output wave file volume level.  A level of 0 to\n"
"                        127 is allowed. (default level is 16)\n"
"\n"
"-t, --turbo             Turbo mode, executes Z80 code as fast as possible.\n"
"                        Without this option the emulation attempts to keep\n"
"                        Z80 CPU execution to match the XTAL value.  This\n"
"                        option is intended for 'hacking' and code development\n"
"                        use.  There are much faster methods if more speed is\n"
"                        required. (see the README file)\n"
"\n"
"--vblank=method         Vertical blanking method to be employed. This is\n"
"                        only intended for 'hacking' when experimenting with\n"
"                        turbo mode and/or high XTAL speeds.  It is not\n"
"                        required or even recommended to be used if 3.375 MHz\n"
"                        or 2 MHz emulation is desired.\n"
"\n"
"                        0 : 50 Hz VBLANK rate derived from Z80 cycles and is\n"
"                            proportional to the XTAL frequency.\n"
"                        1 : 50 Hz VBLANK rate derived from the host timer.\n"
"\n"
"                        When running in turbo mode then setting the <method>\n"
"                        equal to 1 will ensure that a VBLANK rate of 50Hz\n"
"                        will be used.  Without this, key repeating may be\n"
"                        too fast.\n"
"\n"
"--vdu=n                 Premium model option that sets size of VDU RAM to be\n"
"                        emulated.  n may be 2 or 8.  The VDU RAM size\n"
"                        determines the number of screen, attribute and colour\n"
"                        RAM banks.  Standard premium models are 2K.  Default\n"
"                        value is 2K.\n"
"\n"
"--version               Obtain the version number.\n"
"\n"
"--vol=level             Set the sound volume level.  A level of 0 to 127 is\n"
"                        allowed. (default level is 16)\n"
"\n"
"-x, --xtal=frequency    Set the Z80 clock frequency for emulation in MHz.\n"
"                        Standard emulation frequencies are 3.375 and 2.0\n"
"                        MHz.  All other frequencies are classed as 'hacking'.\n" 
"                        3.375 MHz is the default clock frequency.  Frequency\n"
"                        is entered as a floating point number.\n"
"\n"
"                        The highest frequency possible is determined by the\n"
"                        host platform.  As the value increases to the point\n"
"                        where uBee512 can no longer regulate the Z80\n"
"                        execution rate the frame rate will decrease (slower\n"
"                        screen update periods).\n"
"\n"
"-z, --debug             Debugging output (for development purposes only)\n"
"                        writes to a log file.\n"
"\n"
"If you have any new feature suggestions, bug reports, etc. then please send\n"
"an email to: ubee512@gmail.com\n";

 console_output();
 printf("%s", usage);
}

//==============================================================================
// console input/output.
//
// Windows:
// Turns on console input/output so that messages can be seen on a text
// console window.  If this is not called then console out will be
// redirected to stdout.txt and stderr.txt files This is set in the SDL main
// section.
//
// Unices:
// Output for now relies on the emulator started from a command line.
//
//   pass: void
// return: void
//==============================================================================
void console_output (void)
{
#ifdef MINGW
 if (! consoleoutput)
    {
     consoleoutput = 1;
     AllocConsole();
     freopen("conin$", "r", stdin);
     freopen("conout$", "w", stdout);
     freopen("conout$", "w", stderr);
    }
#else
#endif
}

//==============================================================================
// Parameter error message.
//
//   pass: char *parameter
//         char *option_arg
// return: void
//==============================================================================
void param_error_mesg (char *parameter, char *option_arg)
{
#define PARMERR_MESG "ubee512: option %s argument of '%s' is not permitted\n" 
 console_output();
 printf(PARMERR_MESG, parameter, option_arg);
}

//==============================================================================
// Main program start point.
//
//   pass: int argc, char *argv[]	Command line arguments
// return: int				0 if no error, 1 if error
//==============================================================================
int main (int argc, char *argv[])
{
 static struct option long_options[] =
    {
// information output
     {"help",       no_argument,       0, 'h'},
     {"debug",      no_argument,       0, 'z'},
     {"version",    no_argument,       0,  3 },
     {"usage",      no_argument,       0, 'h'},
     {"conio",      no_argument,       0, 18 },

// RS232 communications
     {"coms",       required_argument, 0,  9 },
     {"baud",       required_argument, 0, 10 },
     {"baudtx",     required_argument, 0, 11 },
     {"baudrx",     required_argument, 0, 12 },
     {"datab",      required_argument, 0, 13 },
     {"stopb",      required_argument, 0, 14 },

// parallel printer
     {"print",      required_argument, 0, 15 },
     {"printa",     required_argument, 0, 16 },          

// speed
     {"frate",      required_argument, 0,  4 },
     {"turbo",      no_argument,       0, 't'},
     {"vblank",     required_argument, 0,  2 },
     {"xtal",       required_argument, 0, 'x'},

// drives
     {"image_a",    required_argument, 0, 'a'},
     {"image_b",    required_argument, 0, 'b'},
     {"image_c",    required_argument, 0, 'c'},
     {"image_d",    required_argument, 0, 'd'},

// models
     {"standard",   required_argument, 0, 's'},
     {"model",      required_argument, 0, 17 },

// tape
     {"tapei",      required_argument, 0,  5 },
     {"tapeo",      required_argument, 0,  6 },
     {"tapesamp",   required_argument, 0,  7 },
     {"tapevol",    required_argument, 0,  8 },

// display
     {"aspect",     required_argument, 0,  19},
     {"fullscreen", no_argument,       0, 'f'},
     {"monitor",    required_argument, 0, 'm'},
     {"pcg",        required_argument, 0,  20},
     {"vdu",        required_argument, 0,  21},

// sound
     {"sound",      required_argument, 0,  1 },
     {"vol",        required_argument, 0, 'v'},

// RTC
     {"rtc",        required_argument, 0,  22},

// control
     {"mmode",      no_argument,       0,  23},
     {"nodisk",     no_argument,       0,  24},     

     {0, 	    0,                 0,  0 }
    };

// Do not re-arrange the ordering on any of these values
 char *monitor_args[] =
 {
  "c",
  "a",
  "g",
  "b",
  "w",
  ""
 };

 char *standard_args[] =
 {
  "m",
  "c",  
  ""
 };

 char *sound_args[] =
 {
  "off",
  "prop",
  "normal",
  ""
 };

 int int_arg;
 float float_arg;
 
#define TRY_MESG "%s: Try `%s --help' or `%s --usage' for more information.\n"

 int c;
 int exitstatus = 0;

 while (1)
    {
// getopt_long stores the option index here.
     int option_index = 0;
     
     c = getopt_long (argc, argv, "fhtza:b:c:d:m:s:v:x:", long_options, &option_index);
     
// detect the end of the options.
     if (c == -1)
        break;
     
// do integer and floating point conversions
     if (optarg != NULL)
        {
         int_arg = -1;
         sscanf(optarg, "%d", &int_arg);

         float_arg = -1.0;
         sscanf(optarg, "%f", &float_arg);
        } 
     
     switch (c)
        {
         case  1 : sound = string_search(sound_args, optarg);
                   if (sound == -1) 
                      {
                       param_error_mesg("'--sound'", optarg);
                       exitstatus = 1;
                       break;
                      } 
                   break;
         case  2 : if ((int_arg < 0) || (int_arg > 1))
                      {
                       param_error_mesg("'--vblank'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      vblank_method = int_arg;    
                   break;
         case  3 : console_output();
                   printf("%s\n", APPVER);
                   exitstatus = -1;
                   break;                    
         case  4 : if ((int_arg < 1) || (int_arg > 1000000L))
                      {
                       param_error_mesg("'--frate'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      frame_rate = int_arg;    
                   break;
         case  5 : tape_i = optarg;
                   break;
         case  6 : tape_o = optarg;
                   break;                   
         case  7 : if ((int_arg < 1) || (int_arg > 1000000L))
                      {
                       param_error_mesg("'--tapesamp'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      tape_samples = int_arg;    
                   break;
         case  8 : if ((int_arg < 0) || (int_arg > 127))
                      {
                       param_error_mesg("'--tapevol'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      tape_volume = int_arg;
                   break;   
         case  9 : coms1_dev = optarg;
                   break;
         case 10 : if ((int_arg < 1) || (int_arg > 19200))
                      {
                       param_error_mesg("'--baud'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      {
                       serial_tx_baud = int_arg;
                       serial_rx_baud = int_arg;
                      } 
                   break;
         case 11 : if ((int_arg < 1) || (int_arg > 19200))
                      {
                       param_error_mesg("'--baudtx'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      serial_tx_baud = int_arg;
                   break;
         case 12 : if ((int_arg < 1) || (int_arg > 19200))
                      {
                       param_error_mesg("'--baudrx'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      serial_rx_baud = int_arg;
                   break;
         case 13 : if ((int_arg < 5) || (int_arg > 8))
                      {
                       param_error_mesg("'--datab'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      serial_databits = int_arg;
                   break;
         case 14 : if ((int_arg < 1) || (int_arg > 2))
                      {
                       param_error_mesg("'--stopb'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      serial_stopbits = int_arg;
                   break;
         case 15 : print_b = optarg;
                   break;
         case 16 : print_a = optarg;
                   break;
         case 17 : model = string_search(model_args, optarg);
                   if (model == -1)
                      {
                       param_error_mesg("'--model'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      if (model_config() != 0)
                         exitstatus = 1;
                   break;
         case 18 : console_output();
                   break;
         case 19 : if ((int_arg < 1) || (int_arg > 2))
                      {
                       param_error_mesg("'--aspect'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      opt_aspect = int_arg;
                   break;
         case 20 : if ((int_arg != 16) && (int_arg != 32))
                      {
                       param_error_mesg("'--pcg'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      opt_pcg = int_arg;
                   break;
         case 21 : if ((int_arg != 2) && (int_arg != 8))
                      {
                       param_error_mesg("'--vdu'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      opt_vdu = int_arg;
                   break;
         case 22 : if ((int_arg != 0) && (int_arg != 1))
                      {
                       param_error_mesg("'--rtc'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      rtc_use = int_arg;
                   break;
         case 23 : keyb_force(0x0d, 100);	// force into monitor mode
                   break;
         case 24 : fdc_status0 = 5;
                   break;
         case 'f': fullscreen = 1;
                   break;
         case 'h': ApplicationUsage();
                   exitstatus = -1;
                   break;
         case 't': turbo = 1;
                   break;
         case 'z': debuglvl = 1;
                   break;
         case 'a': floppy_a = optarg;
                   break;
         case 'b': floppy_b = optarg;
                   break;
         case 'c': floppy_c = optarg;
                   break;
         case 'd': floppy_d = optarg;
                   break;
         case 'm': monitor = string_search(monitor_args, optarg);
                   if (monitor == -1)
                      {
                       param_error_mesg("'-m, --monitor'", optarg);
                       exitstatus = 1;
                       break;
                      } 
                   break;
         case 's': colour_use = string_search(standard_args, optarg);
                   if (colour_use != -1)
                      standard = 1;
                   else
                      {
                       param_error_mesg("'-s, --standard'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   break;
         case 'v': int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 0) || (int_arg > 127))
                      {
                       param_error_mesg("'--vol'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      audio_volume = int_arg;
                   break;
         case 'x': if (float_arg <= 0.0)
                      {
                       param_error_mesg("'-x, --xtal'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      opt_xtal = float_arg;    
                   break;
         case '?': // getopt_long already printed an error message.
                   console_output();
#ifdef MINGW
                   // now do it again becuase it would only have gone
                   // to stderr.txt
                   printf("%s: invalid option -- %c\n", argv[0], c);
#endif
                   printf(TRY_MESG, argv[0], argv[0], argv[0]);
                   exitstatus = 1;
                   break;
         default : exitstatus = 1;
                   break;
        }
    }

// Check if any other arguments are incorrectly specified
 if (optind < argc)
    {
     console_output();
     exitstatus = 1;
     printf("%s: %d additional arguments were specified that are not recognised:\n", argv[0], argc - optind);         
     printf("%s: ", argv[0]);
     while (optind < argc)
        printf ("%s ", argv[optind++]);
     putchar('\n');
     printf(TRY_MESG, argv[0], argv[0], argv[0]);
    }

// if no option errors then initialise the emulator
 if (exitstatus == 0)
    {
     if (init() != 0)
        {
         console_output();
         exitstatus = 1;
         printf("main: Fatal error during initialisation.\n");
         SDL_Quit();
        }
    }

// if no errors then start the emulator
 if (exitstatus == 0)
    {
     ApplicationSetup();
     ApplicationLoop();
    }

// if no errors then de-initialise the emulator
 if (exitstatus == 0)
    {
     if (deinit() != 0)
        {
         console_output();
         exitstatus = 1;
         printf("main: Error while de-initialising.\n");
        }
    }    

// if running on Windows then get a confirmation before the console
// output window is closed.
 if (exitstatus)
    {
#ifdef MINGW
     gui_message_box(BUTTON_OK, "Read message(s) in console output window before closing.");
#endif
     if (exitstatus == -1)
        return 0;
     else   
        return exitstatus;
    }

 return 0;
}
